#!/usr/bin/env python3
#
# Copyright      2023  Xiaomi Corp.       (authors: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R match_test_py

import unittest
import numpy as np

from textsearch import get_longest_increasing_pairs


class TestMatch(unittest.TestCase):
    def test_get_longest_increasing_pairs(self):
        for dtype in [np.int32, np.int64]:
            seq1 = np.array([0, 1, 1, 2, 2, 3, 4, 5, 6], dtype=dtype)
            seq2 = np.array([9, 7, 8, 9, 6, 7, 10, 12, 8], dtype=dtype)
            expected = [(1, 7), (1, 8), (2, 9), (4, 10), (5, 12)]
            result = get_longest_increasing_pairs(seq1=seq1, seq2=seq2)
            for i, r in enumerate(result):
                self.assertTrue(r == expected[i])


if __name__ == "__main__":
    unittest.main()



# Copyright      2023   Xiaomi Corp.       (author: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import _textsearch
import numpy as np


def _renumbering(array: np.ndarray) -> np.ndarray:
    """Renumber element in the input array such that the returned array
    contains entries ranging from 0 to M - 1, where M equals
    to number of unique entries in the input array.

    The order of entries in the output array is the same as the order
    of entries in the input array. That is, if array[i] < array[j], then
    ans[i] < ans[j].

    Args:
      array:
        A 1-D array.
    Returns:
      Return a renumbered 1-D array.
    """
    uniqued, inverse = np.unique(array, return_inverse=True)
    # Note: uniqued[inverse] == array

    indexes_sorted2unsorted = np.argsort(uniqued)
    indexes_unsorted2sorted = np.empty((uniqued.size), dtype=np.int32)
    indexes_unsorted2sorted[indexes_sorted2unsorted] = np.arange(uniqued.size)

    return indexes_unsorted2sorted[inverse]


def create_suffix_array(array: np.ndarray) -> np.ndarray:
    """Create a suffix array from a 1-D input array.

    hint:
      Please refer to https://en.wikipedia.org/wiki/Suffix_array
      for what suffix array is. Different from the above Wikipedia
      article the special sentinel letter ``$`` in `textsearch`_
      is known as EOS and it is larger than any other characters.

    Args:
      array:
        A 1-D integer (or unsigned integer) array of shape ``(seq_len - 1,)``.

        Note:
          Inside this function, we will append explicitly an EOS
          symbol that is larger than ``array.max()``.
    Returns:
      Returns a suffix array of type ``np.int32``, of shape ``(seq_len,)``.
      This will consist of some permutation of the elements
      ``0 .. seq_len - 1``.

    **Usage examples**:

        .. literalinclude:: code/suffix-array.py
    """
    assert array.ndim == 1, array.ndim

    # Renumber elements in the array so that array.max() equals
    # to the number of unique elements in the array.
    #
    # In the implementation, we allocate an array of size array.max().
    # A smaller value of array.max() leads to less memory allocation.
    array = _renumbering(array)

    max_symbol = array.max()
    assert max_symbol < np.iinfo(array.dtype).max - 1, max_symbol
    eos = max_symbol + 1
    padding = np.array([eos, 0, 0, 0], dtype=array.dtype)

    padded_array = np.concatenate([array, padding])

    # The C++ code requires the input array to be contiguous.
    array_int32 = np.ascontiguousarray(padded_array, dtype=np.int32)
    return _textsearch.create_suffix_array(array_int32)



#!/usr/bin/env python3
#
# Copyright      2024  Xiaomi Corp.       (authors: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R match_test_py

import unittest

from textsearch.utils import is_overlap


class TestOverlap(unittest.TestCase):
    def test_is_overlap(self):
        candidates = [
            [20, 30],
            [15, 25],
            [10, 21.1],
            [1, 10],
            [60, 70],
            [65, 73],
            [68.5, 85],
            [25, 35],
            [45, 55],
            [20, 25],
            [21, 25],
            [34.5, 46.5],
            [35, 46.1],
            [25, 35],
            [26, 34],
            [44, 70.5],
        ]
        selected_ranges: List[Tuple[float, float]] = []
        selected_indexes: List[int] = []
        segments = []
        overlapped_segments = []
        for r in candidates:
            status, index = is_overlap(
                selected_ranges,
                selected_indexes,
                query=(r[0], r[1]),
                segment_index=len(segments),
                overlap_ratio=0.1,
            )
            if status:
                if index is not None:
                    overlapped_segments.append(index)
                    segments.append(r)
            else:
                segments.append(r)
        for index in sorted(overlapped_segments, reverse=True):
            segments.pop(index)
        expected_segments = [
            [10, 21.1],
            [1, 10],
            [68.5, 85],
            [25, 35],
            [21, 25],
            [35, 46.1],
        ]
        assert segments == expected_segments


if __name__ == "__main__":
    unittest.main()



import json
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Callable, List, Optional, Union

from .utils import row_ids_to_row_splits

import numpy as np


@dataclass
class TextSource:
    """
    Represents the full text of a utf-8 document.

    The user can choose whether to save it in a utf-8 encoded np.uint8 array
    or to save its Unicode codepoint into a np.int32 array.
    """

    # filename or ID
    name: str

    # It can be either a np.uint8 array or a np.int32 array.
    # If it is a np.uint8 array, it contains the utf-8 encoded bytes;
    # If it is a np.int32 array, it contains the Unicode codepoint of the text.
    # Length of the array may not be >= 2**32.
    binary_text: np.ndarray

    # Only used when binary_text.dtype is np.int32.
    # It contains the mapping from utf-8 character position to byte position.
    # That is pos[i] contains the byte position in binary_text for the i-th
    # utf-8 character
    pos: Optional[np.ndarray] = None

    # Whether has punctuation in this TextSource, if True we will only
    # break the query at punctuation position when splitting aligned segment
    # into smaller pieces.
    has_punctuation: bool = False

    @property
    def text(self) -> str:
        """Return Python string representation of self.binary_text decoded as UTF-8."""
        if self.binary_text.dtype == np.uint8:
            return self.binary_text.tobytes().decode("utf-8")
        else:
            assert self.binary_text.dtype == np.int32, self.binary_text.dtype
            return "".join([chr(i) for i in self.binary_text])

    @staticmethod
    def from_str(
        name: str, s: str, use_utf8: bool, has_punctuation: bool = False
    ) -> "TextSource":
        """Construct an instance of TextSource from a string.

        Args:
          name:
            Name of the returned instance. It can be either a filename or an ID.
          s:
            It contains the text string.
          use_utf8:
            True to encode the text with utf-8.
            False to save the Unicode codepoint of the text.
        """

        if use_utf8:
            binary_text = s.encode("utf-8")
            return TextSource(
                name=name,
                binary_text=np.frombuffer(binary_text, dtype=np.uint8),
                pos=None,
                has_punctuation=has_punctuation,
            )
        else:
            binary_text = np.fromiter(
                (ord(i) for i in s), dtype=np.int32, count=len(s)
            )
            pos = _find_byte_offsets_for_utf8_symbols(binary_text)

            return TextSource(
                name=name,
                binary_text=binary_text,
                pos=pos,
                has_punctuation=has_punctuation,
            )


def _find_byte_offsets_for_utf8_symbols(binary_text: np.ndarray) -> np.ndarray:
    """
    Args:
      binary_text:
        A 1-D int32 array containing Unicode codepoint of characters.
    Returns:
      Return a np.uint32 1-D array. Its length equals to the number of
      characters contained in binary_text before encoding. ans[i]
      contains the byte position in binary_text for the i-th character.
    """
    byte_offsets = []

    # Iterate over each byte in the binary text
    byte_index = 0
    for i in binary_text:
        if i < 0x80:
            # Single-byte UTF-8 symbol
            symbol_length = 1
        elif i < 0x800:
            # Two-byte UTF-8 symbol
            symbol_length = 2
        elif i < 0x10000:
            # Three-byte UTF-8 symbol
            symbol_length = 3
        else:
            # Four-byte UTF-8 symbol
            symbol_length = 4

        # Add the byte index to the list of offsets
        byte_offsets.append(byte_index)

        # Move the byte index to the start of the next symbol
        byte_index += symbol_length

    # Note: uint32 should be sufficient for 4GB of text,
    # I don't expect we'd deal with such large documents here.
    return np.asarray(byte_offsets, dtype=np.uint32)


@dataclass
class Transcript:
    """
    Type that represents automatically transcribed text.
    The time marks are at the byte level.  We can easily turn automatic
    transcripts of text into this type as long as we keep track of the time for each symbol.
    """

    # A filename or an ID.
    name: str

    # the text as a sequence of bytes or (more likely) int32 corresponding to UTF codepoints.
    # We will have to expand BPE tokens into bytes, presumably converting _ into space.
    #
    # When its dtype is np.uint8, it is encoded with utf-8
    # When its dtype is np.int32, it contains the Unicode code point of characters
    binary_text: np.ndarray

    # Gives the time in seconds for each byte of the text.  (Should be in non-decreasing order).
    # Its dtype is np.float32
    times: np.ndarray

    @property
    def text(self) -> str:
        """Return Python string representation of self.binary_text decoded as UTF-8."""
        if self.binary_text.dtype == np.uint8:
            return self.binary_text.tobytes().decode("utf-8")
        else:
            assert self.binary_text.dtype == np.int32, self.binary_text.dtype
            return "".join([chr(i) for i in self.binary_text])

    @staticmethod
    def from_dict(
        name: str, d: dict, use_utf8: bool, is_bpe: bool = False
    ) -> "Transcript":
        """
        Args:
          name:
            A filename or an ID.
          d:
            A dict containing:

              - d["text"]: List[str]. Each element in the list is a token or a word.

              - d["begin_times]: List[float]. len(d["text"]) == d["begin_times"].
                d["begin_times"][i] is the begin time for d["text"][i]
          use_utf8:
            True to use utf-8 to encode the text.
            False to save the Unicode code point of the text.
        """
        assert "text" in d, list(d.keys())
        assert "begin_times" in d, list(d.keys())
        assert len(d["text"]) == len(d["begin_times"]), (
            len(d["text"]),
            len(d["begin_times"]),
            d["text"],
            d["begin_times"],
        )

        if use_utf8:
            byte_list = []
            time_list = []
            for text, begin_time in zip(d["text"], d["begin_times"]):
                text = text.replace("▁", " ") if is_bpe else text

                b = text.encode("utf-8")

                if time_list:
                    # Check that begin_time is non-decreasing.
                    #
                    # < here requires that it is strictly increasing.
                    assert time_list[-1] < begin_time, (
                        time_list[-1],
                        begin_time,
                    )

                # bytes belonging to the same text have the same begin time
                time_list += [begin_time] * len(b)
                byte_list.append(b)

            return Transcript(
                name=name,
                binary_text=np.frombuffer(b"".join(byte_list), dtype=np.uint8),
                times=np.asarray(time_list, dtype=np.float32),
            )
        else:
            codepoint_list = []
            time_list = []
            for text, begin_time in zip(d["text"], d["begin_times"]):
                text = text.replace("▁", " ") if is_bpe else text

                codepoint_list.append(ord(i) for i in text)

                if time_list:
                    # Check that begin_time is non-decreasing.
                    #
                    # < here requires that it is strictly increasing.
                    assert time_list[-1] < begin_time, (
                        time_list[-1],
                        begin_time,
                    )

                # bytes belonging to the same text have the same begin time
                # Each character occupies 4 bytes, so it is multiplied by 4
                time_list += [begin_time] * (len(text) * 4)

            return Transcript(
                name=name,
                binary_text=np.fromiter(chain(*codepoint_list), dtype=np.int32),
                times=np.asarray(time_list, dtype=np.float32),
            )


# we'll have a global list of text sources during the program lifetime, and indexes into this list
# (probably int32) will represent the text source.
TextSources = List[Union[TextSource, Transcript]]


@dataclass
class SourcedText:
    """
    Represents a text with some meta-info that records from where in a collection of
    texts it came.
    """

    # A 1-d array. Its data type can be either np.uint8 or np.int32.
    # If it is np.uint8, it represents utf-8 encoded strings.
    # If it is np.int32, it represents strings with Unicode code point.
    #
    # Note: this text array is 'owned' here so it can be modified by the user.
    binary_text: np.ndarray

    # The text position in the original source for each element of the text. Of type uint32;
    pos: np.ndarray

    # The document index for each element of the text.
    # The np.ndarray would be of type uint32.
    doc: np.ndarray

    # For reference, the list of available text sources that this text might come from.
    sources: TextSources

    # Optional: an array derived from `doc` which is like the row
    # splits in k2 (or in tensorflow ragged arrays), with `doc` being
    # the row-ids.
    _doc_splits: Optional[np.ndarray] = None

    @property
    def doc_splits(self) -> np.ndarray:
        assert isinstance(self.doc, np.ndarray), type(self.doc)
        if self._doc_splits is not None:
            return self._doc_splits

        self._doc_splits = row_ids_to_row_splits(self.doc)

        return self._doc_splits


def texts_to_sourced_texts(
    sources: TextSources, uppercase: bool = False
) -> List[SourcedText]:
    """Construct a list of SourcedText from TextSources.
    We have len(ans) == len(sources).
    """
    ans = []

    for s in sources:
        # pos[i] contains the index of the i-th entry in s.binary_text
        #
        # If s.binary_text.dtype is np.int32, pos[i] is the position of the i-th
        # Unicode code point in s.binary_text
        #
        # If s.binary_text.dtype is np.uint8, pos[i] is the position of the
        # i-th byte of s.binary_text
        pos = np.arange(s.binary_text.size, dtype=np.uint32)
        doc = np.zeros(s.binary_text.size, dtype=np.uint32)

        binary_text = s.binary_text
        if uppercase:
            # 32 = ord('a')- ord('A')
            # 97 = ord('a')
            # 122 = ord('z')
            binary_text = np.where(
                ((binary_text >= 97) & (binary_text <= 122)),
                binary_text - 32,
                binary_text,
            )

        ans.append(
            SourcedText(
                binary_text=binary_text,
                pos=pos,
                doc=doc,  # indexes into the following `sources` attribte
                sources=[s],
            )
        )

    return ans


def append_texts(texts: List[SourcedText]) -> SourcedText:
    if len(texts) == 1:
        # return a shallow copy
        return texts[0]

    # Check that all texts have the same dtype
    for t in texts[1:]:
        assert t.binary_text.dtype in (np.uint8, np.int32), t.dtype
        assert texts[0].binary_text.dtype == t.binary_text.dtype, (
            texts[0].binary_text.dtype,
            t.binary_text.dtype,
        )

    binary_text = np.concatenate([t.binary_text for t in texts])
    pos = np.concatenate([t.pos for t in texts])

    doc_list = []
    for t in texts:
        if isinstance(t.doc, np.ndarray):
            assert t.doc.dtype == np.uint32, t.doc.dtype
            doc_list.append(t.doc)
        else:
            assert isinstance(t.doc, int), type(t.doc)
            doc_list.append(np.zeros(t.binary_text.size, dtype=np.uint32))

    num_docs = 0
    for i, d in enumerate(texts):
        if i == 0:
            num_docs = len(d.sources)
            continue
        doc_list[i] += num_docs
        num_docs += len(d.sources)

    doc = np.concatenate(doc_list)
    assert doc.dtype == np.uint32, doc.dtype

    sources = [s for t in texts for s in t.sources]

    return SourcedText(
        binary_text=binary_text, pos=pos, doc=doc, sources=sources,
    )


def filter_texts(
    t: SourcedText,
    fn: Optional[Callable[[Union[np.int32, np.uint8]], bool]] = None,
    keep: Optional[np.ndarray] = None,
) -> SourcedText:
    """
    Filter some elements from a SourcedText (out-of-place).
    Args:
        t:
          the text to remove some positions of
        fn:
          a function takes each element as input and output whether to keep this element.
        keep:
          an np.ndarray with dtype == np.bool, that is True
          for positions that should be kept; must have the same shape
          as t.text.
    Returns:
        A SourcedText with some positions removed.
    """
    # TODO: Checking if this also works correctly for utf8.
    assert t.binary_text.dtype == np.int32
    if keep is None:
        assert fn is not None
        vfn = np.vectorize(fn)
        keep = vfn(t.binary_text)
    assert keep.ndim == 1, keep.ndim
    new2old = keep.nonzero()[0]
    binary_text = t.binary_text[new2old]
    pos = t.pos[new2old]
    doc = t.doc
    if not isinstance(t.doc, int):
        doc = t.doc[new2old]
    return SourcedText(
        binary_text=binary_text, pos=pos, doc=doc, sources=t.sources,
    )



# Copyright      2023   Xiaomi Corp.       (author: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from typing import List, Tuple


def get_nice_alignments(
    alignments: List[Tuple[int, int, str]],
    query: np.ndarray,
    target: np.ndarray,
) -> List[str]:
    """
    Get the alignment of the matched segments.

    Args:
      alignments:
        The alignments information returned by the :func:`levenshtein_distance`.
      query:
        The query sequence.
      target:
        The target sequence.

    Returns:
      Return a list of alignment string, it has the same length as alignments.
      See the following example for more details, the first line of the alignment
      is query, the second line is the error types, the third line is the target
      segment. The following symbols are used to represent the error types:

        - ``*``: empty
        - ``+``: insertion
        - ``-``: deletion
        - ``#``: substitution
        - ``|``: correct.

    **Example 1/2**

    >>> from textsearch import get_nice_alignments, levenshtein_distance
    >>> import numpy as np
    >>> query = np.array([1, 2, 3, 4], dtype=np.int32)
    >>> target = np.array([1, 5, 3, 4, 6, 7, 1, 2, 4], dtype=np.int32)
    >>> distance, alignments = levenshtein_distance(query, target)
    >>> distance
    1
    >>> alignments
    [(0, 3, 'CSCC'), (6, 8, 'CCIC')]
    >>> aligns = get_nice_alignments(alignments, query, target)
    >>> repr(aligns)
    "['1 2 3 4 \\n| # | | \\n1 5 3 4 ', '1 2 3 4 \\n| | + | \\n1 2 * 4 ']"
    >>> print (aligns[0])
    1 2 3 4
    | # | |
    1 5 3 4
    >>> print (aligns[1])
    1 2 3 4
    | | + |
    1 2 * 4

    **Example 2/2**

        .. literalinclude:: ./code/edit-distance.py
    """
    results = []
    for align in alignments:
        j = align[0]
        i = 0
        ali_seq = align[2]
        qs = ""
        ms = ""
        ts = ""
        for k in ali_seq:
            if k == "D":
                # a deletion error
                ts_ = f"{target[j]} "
                sl = len(ts_)

                qs += f"{'*':{sl}}"
                ms += f"{'-':{sl}}"
                ts += f"{ts_:{sl}}"
                j = j + 1
            elif k == "S" or k == "C":
                # correct or a substitution error
                qs_ = f"{query[i]} "
                ts_ = f"{target[j]} "
                sl = max(len(qs_), len(ts_))

                qs += f"{qs_:{sl}}"
                ts += f"{ts_:{sl}}"
                if k == "C":
                    ms += f"{'|':{sl}}"
                else:
                    ms += f"{'#':{sl}}"
                j = j + 1
                i = i + 1
            else:
                # an insertion error
                assert k == "I", k
                qs_ = f"{query[i]} "
                sl = len(qs_)

                qs += f"{qs_:{sl}}"
                ms += f"{'+':{sl}}"
                ts += f"{'*':{sl}}"
                i = i + 1
        results.append("\n".join([qs, ms, ts]))
    return results



#!/usr/bin/env python3

import unittest

import numpy as np

from textsearch import SourcedText
from textsearch import TextSource
from textsearch import Transcript
from textsearch import append_texts
from textsearch import filter_texts
from textsearch import texts_to_sourced_texts


class TestSourcedText(unittest.TestCase):
    def test_from_text_source_utf8(self):
        """Test constructing a SourcedText from a TextSource using utf-8"""
        name = "test_text_source_utf8"
        s = "zażółć gęślą jaźń\n你好Hallo"
        text_source = TextSource.from_str(name=name, s=s, use_utf8=True)

        sourced_text = texts_to_sourced_texts([text_source])
        assert len(sourced_text) == 1, len(sourced_text)
        assert sourced_text[0].pos.dtype == np.uint32, sourced_text[0].pos.dtype
        np.testing.assert_equal(
            sourced_text[0].pos, np.arange(text_source.binary_text.size)
        )
        np.testing.assert_equal(sourced_text[0].doc,np.zeros(text_source.binary_text.size))
        assert len(sourced_text[0].sources) == 1, len(sourced_text[0].sources)
        assert sourced_text[0].sources[0] is text_source

    def test_from_text_source_unicode_code_point(self):
        """Test constructing a SourcedText from a TextSource using Unicode code point"""
        name = "test_text_source_unicode_code_point"
        s = "zażółć gęślą jaźń\n你好Hallo"
        text_source = TextSource.from_str(name=name, s=s, use_utf8=False)

        sourced_text = texts_to_sourced_texts([text_source])
        assert len(sourced_text) == 1, len(sourced_text)
        assert sourced_text[0].pos.dtype == np.uint32, sourced_text[0].pos.dtype
        np.testing.assert_equal(
            sourced_text[0].pos, np.arange(text_source.binary_text.size)
        )
        np.testing.assert_equal(sourced_text[0].doc, np.zeros(text_source.binary_text.size))
        assert len(sourced_text[0].sources) == 1, len(sourced_text[0].sources)
        assert sourced_text[0].sources[0] is text_source

    def test_from_transcript_utf8(self):
        """Test constructing a SourcedText from a Transcript using utf-8"""
        name = "test_from_transcript_utf8"
        text = "▁I ▁saw ▁a ▁girl ▁with ▁a ▁ te le s c o pe".split()
        begin_times = []
        for i in range(len(text)):
            begin_times.append(i * 0.25)
        d = {
            "text": text,
            "begin_times": begin_times,
        }

        transcript = Transcript.from_dict(name=name, d=d, use_utf8=True)

        sourced_text = texts_to_sourced_texts([transcript])
        assert len(sourced_text) == 1, len(sourced_text)
        assert sourced_text[0].pos.dtype == np.uint32, sourced_text[0].pos.dtype
        np.testing.assert_equal(
            sourced_text[0].pos, np.arange(transcript.binary_text.size)
        )

        np.testing.assert_equal( sourced_text[0].doc, np.zeros(transcript.binary_text.size))
        assert len(sourced_text[0].sources) == 1, len(sourced_text[0].sources)
        assert sourced_text[0].sources[0] is transcript

    def test_from_transcript_unicode_code_point(self):
        """Test constructing a SourcedText from a Transcript using Unicode code point"""
        name = "test_from_transcript_unicode_code_point"
        text = "▁I ▁saw ▁a ▁girl ▁with ▁a ▁ te le s c o pe".split()
        begin_times = []
        for i in range(len(text)):
            begin_times.append(i * 0.25)
        d = {
            "text": text,
            "begin_times": begin_times,
        }

        transcript = Transcript.from_dict(name=name, d=d, use_utf8=False)

        sourced_text = texts_to_sourced_texts([transcript])
        assert len(sourced_text) == 1, len(sourced_text)
        assert sourced_text[0].pos.dtype == np.uint32, sourced_text[0].pos.dtype
        np.testing.assert_equal(
            sourced_text[0].pos, np.arange(transcript.binary_text.size)
        )
        np.testing.assert_equal(sourced_text[0].doc, np.zeros(transcript.binary_text.size))
        assert len(sourced_text[0].sources) == 1, len(sourced_text[0].sources)
        assert sourced_text[0].sources[0] is transcript

    def test_append_texts_text_source_utf8(self):
        name0 = "test_append_texts_text_source_utf8_0"
        s0 = "zażółć gęślą jaźń\n"
        text_source0 = TextSource.from_str(name=name0, s=s0, use_utf8=True)

        name1 = "test_append_texts_text_source_utf8_1"
        s1 = "你好Hallo"
        text_source1 = TextSource.from_str(name=name1, s=s1, use_utf8=True)

        sourced_text0 = texts_to_sourced_texts([text_source0])
        sourced_text1 = texts_to_sourced_texts([text_source1])
        sourced_text = append_texts([sourced_text0[0], sourced_text1[0]])

        # check binary_text
        expected_binary_text = np.concatenate(
            [text_source0.binary_text, text_source1.binary_text]
        )
        np.testing.assert_equal(sourced_text.binary_text, expected_binary_text), (
            sourced_text.binary_text,
            expected_binary_text,
        )
        assert (
            sourced_text.binary_text.dtype == np.uint8
        ), sourced_text.binary_text.dtype

        # check pos
        expected_pos0 = np.arange(text_source0.binary_text.size, dtype=np.uint32)
        expected_pos1 = np.arange(text_source1.binary_text.size, dtype=np.uint32)
        expected_pos = np.concatenate([expected_pos0, expected_pos1])
        np.testing.assert_equal(sourced_text.pos, expected_pos), (
            sourced_text.pos,
            expected_pos,
        )
        assert sourced_text.pos.dtype == np.uint32, sourced_text.pos.dtype

        # check doc
        expected_doc0 = np.zeros(text_source0.binary_text.size, dtype=np.uint32)
        expected_doc1 = np.ones(text_source1.binary_text.size, dtype=np.uint32)
        expected_doc = np.concatenate([expected_doc0, expected_doc1])
        np.testing.assert_equal(sourced_text.doc, expected_doc), (
            sourced_text.doc,
            expected_doc,
        )
        assert sourced_text.doc.dtype == np.uint32, sourced_text.doc.dtype

        # check sources
        assert len(sourced_text.sources) == 2, len(sourced_text.sources)
        assert sourced_text.sources[0] is text_source0
        assert sourced_text.sources[1] is text_source1

        expected_doc_splits = np.array(
            [
                0,
                text_source0.binary_text.size,
                text_source0.binary_text.size + text_source1.binary_text.size,
            ],
            dtype=np.uint32,
        )
        np.testing.assert_equal(sourced_text.doc_splits, expected_doc_splits)
        assert sourced_text.doc_splits.dtype == np.uint32, sourced_text.doc_splits.dtype

    def test_append_texts_text_source_unicode_code_point(self):
        name0 = "test_append_texts_text_source_unicode_code_point_0"
        s0 = "zażółć gęślą jaźń\n"
        text_source0 = TextSource.from_str(name=name0, s=s0, use_utf8=False)

        name1 = "test_append_texts_text_source_unicode_code_point_1"
        s1 = "你好Hallo"
        text_source1 = TextSource.from_str(name=name1, s=s1, use_utf8=False)

        sourced_text0 = texts_to_sourced_texts([text_source0])
        sourced_text1 = texts_to_sourced_texts([text_source1])
        sourced_text = append_texts([sourced_text0[0], sourced_text1[0]])

        # check binary_text
        expected_binary_text = np.concatenate(
            [text_source0.binary_text, text_source1.binary_text]
        )
        np.testing.assert_equal(sourced_text.binary_text, expected_binary_text), (
            sourced_text.binary_text,
            expected_binary_text,
        )
        assert (
            sourced_text.binary_text.dtype == np.int32
        ), sourced_text.binary_text.dtype

        # check pos
        expected_pos0 = np.arange(text_source0.binary_text.size, dtype=np.uint32)
        expected_pos1 = np.arange(text_source1.binary_text.size, dtype=np.uint32)
        expected_pos = np.concatenate([expected_pos0, expected_pos1])
        np.testing.assert_equal(sourced_text.pos, expected_pos), (
            sourced_text.pos,
            expected_pos,
        )
        assert sourced_text.pos.dtype == np.uint32, sourced_text.pos.dtype

        # check doc
        expected_doc0 = np.zeros(text_source0.binary_text.size, dtype=np.uint32)
        expected_doc1 = np.ones(text_source1.binary_text.size, dtype=np.uint32)
        expected_doc = np.concatenate([expected_doc0, expected_doc1])
        np.testing.assert_equal(sourced_text.doc, expected_doc), (
            sourced_text.doc,
            expected_doc,
        )
        assert sourced_text.doc.dtype == np.uint32, sourced_text.doc.dtype

        # check sources
        assert len(sourced_text.sources) == 2, len(sourced_text.sources)
        assert sourced_text.sources[0] is text_source0
        assert sourced_text.sources[1] is text_source1

        expected_doc_splits = np.array(
            [
                0,
                text_source0.binary_text.size,
                text_source0.binary_text.size + text_source1.binary_text.size,
            ],
            dtype=np.uint32,
        )
        np.testing.assert_equal(sourced_text.doc_splits, expected_doc_splits)
        assert sourced_text.doc_splits.dtype == np.uint32, sourced_text.doc_splits.dtype

    def test_filter_texts(self):
        name0 = "test_filter_texts_0"
        s0 = "Higher, faster, stronger, together"
        text_source0 = TextSource.from_str(name=name0, s=s0, use_utf8=False)

        sourced_text0 = texts_to_sourced_texts([text_source0])
        sourced_text = append_texts([sourced_text0[0]])

        # test keep
        keep = np.array([True if x != "," else False for x in s0], dtype=bool)
        new_sourced_text = filter_texts(sourced_text, keep=keep)

        new_s0 = "Higher faster stronger together"
        new_text_source0 = TextSource.from_str(name=name0, s=new_s0, use_utf8=False)
        np.testing.assert_equal(
            new_sourced_text.binary_text, new_text_source0.binary_text
        )

        new_pos_list = list(range(len(s0)))
        new_pos_list.remove(6)
        new_pos_list.remove(14)
        new_pos_list.remove(24)
        new_pos = np.array(new_pos_list, dtype=np.uint32)
        np.testing.assert_equal(new_sourced_text.pos, new_pos)

        # test fn
        fn = lambda x: x != ord(",")
        new_sourced_text = filter_texts(sourced_text, fn=fn)
        np.testing.assert_equal(
            new_sourced_text.binary_text, new_text_source0.binary_text
        )
        np.testing.assert_equal(new_sourced_text.pos, new_pos)


if __name__ == "__main__":
    unittest.main()



from _textsearch import find_close_matches
from _textsearch import levenshtein_distance

from .datatypes import SourcedText
from .datatypes import TextSource
from .datatypes import TextSources
from .datatypes import Transcript
from .datatypes import append_texts
from .datatypes import filter_texts
from .datatypes import texts_to_sourced_texts

from .levenshtein import get_nice_alignments

from .match import align_queries
from .match import get_longest_increasing_pairs
from .match import split_aligned_queries

from .suffix_array import create_suffix_array

from .utils import AttributeDict
from .utils import is_punctuation
from .utils import row_ids_to_row_splits
from .utils import str2bool
__version__ = '0.10'



#!/usr/bin/env python3
#
# Copyright      2023  Xiaomi Corp.       (authors: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R suffix_array_test_py

import unittest
import numpy as np

from textsearch import create_suffix_array


class TestSuffixArray(unittest.TestCase):
    def test_create_suffix_array(self):
        for dtype in [np.uint8, np.int8, np.uint16, np.int16]:
            array = np.array([3, 2, 1], dtype=dtype)
            suffix_array = create_suffix_array(array)
            expected_array = np.array([2, 1, 0, 3], dtype=np.int32)
            self.assertTrue((suffix_array == expected_array).all())
            self.assertTrue(suffix_array.dtype == np.int32)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
#
# Copyright      2023  Xiaomi Corp.       (authors: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R levenshtein_distance_test_py

import unittest
import numpy as np

from textsearch import get_nice_alignments, levenshtein_distance


class TestLevenshtein(unittest.TestCase):
    def test_levenshtein_distance(self):
        query = np.array([1, 2, 3, 4], dtype=np.int32)
        target = np.array([1, 5, 3, 4, 6, 7, 1, 2, 4], dtype=np.int32)
        distance, alignments = levenshtein_distance(query, target)
        self.assertTrue(distance == 1)
        self.assertTrue(len(alignments) == 2)
        self.assertTrue(alignments[0] == (0, 3, "CSCC"))
        self.assertTrue(alignments[1] == (6, 8, "CCIC"))

    def test_get_nice_alignments(self):
        query = np.array([10, 234, 98745, 14, 8], dtype=np.int32)
        target = np.array([7, 10, 134, 9, 98745, 8], dtype=np.int32)
        distance, alignments = levenshtein_distance(query, target)
        align_str = get_nice_alignments(alignments, query, target)
        expected_align = (
            "10 234 * 98745 14 8 \n|  #   - |     +  | \n10 134 9 98745 *  8 "
        )
        self.assertTrue(len(align_str) == 1)
        self.assertTrue(align_str[0] == expected_align)


if __name__ == "__main__":
    unittest.main()



# Copyright      2023   Xiaomi Corp.       (author: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
from bisect import bisect_left, bisect_right
from dataclasses import dataclass
from heapq import heappush, heappop
from multiprocessing.pool import Pool
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import regex

from _textsearch import (
    find_close_matches,
    get_longest_increasing_pairs as _get_longest_increasing_pairs,
    levenshtein_distance,
)
from .suffix_array import create_suffix_array
from .datatypes import SourcedText, TextSource, Transcript
from .utils import (
    PUCTUATIONS,
    is_overlap,
    is_punctuation,
    row_ids_to_row_splits,
)


def get_longest_increasing_pairs(
    seq1: np.ndarray, seq2: np.ndarray
) -> List[Tuple[int, int]]:
    """
    Get the longest increasing pairs for given sequences.
    See https://github.com/danpovey/text_search/issues/21 for more details.

    Suppose seq1 is [i1, i2, i3... iN] and seq2 is [j1, j2, j3... jN], this
    function returns the  longest increasing pairs: (i1, j1), (i2, j2), ... (iN, jN)
    such that i1 <= i2 <= ... <= iN, and j1 <= j2 <= ... <= jN.

    Args:
      seq1:
        The first sequence.
      seq2:
        The second sequence.

    >>> import numpy as np
    >>> from textsearch import get_longest_increasing_pairs
    >>> seq1 = np.array([0, 1, 1, 2, 2, 3, 4, 5, 6], dtype=np.int32)
    >>> seq2 = np.array([9, 7, 8, 9, 6, 7, 10, 12, 8], dtype=np.int64)
    >>> get_longest_increasing_pairs(seq1=seq1, seq2=seq2)
    [(1, 7), (1, 8), (2, 9), (4, 10), (5, 12)]
    """
    assert seq1.ndim == 1, seq1.ndim
    assert seq2.ndim == 1, seq2.ndim
    assert seq1.size == seq2.size, (seq1.size, seq2.size)

    # The sequences are required to be contiguous int32 array in C++.
    seq1_int32 = np.ascontiguousarray(seq1, dtype=np.int32)
    seq2_int32 = np.ascontiguousarray(seq2, dtype=np.int32)

    return _get_longest_increasing_pairs(seq1_int32, seq2_int32)


def _break_query(
    sourced_text: SourcedText,
    matched_points: List[Tuple[int, int]],
    segment_length: int = 5000,
    reference_length_difference: float = 0.1,
) -> List[Tuple[int, int, int, int]]:
    """
    Break long query into small segments at the matched points with the
    `segment_length` constraint.

    Args:
      sourced_text:
        The SourcedText containing the queries and references.
      matched_points:
        A list of matched points, each item is a pair of (query index, target index)
        in sourced_text.
      segment_length:
        The expected length of the segmented piece. Note: this is not really the
        length of the segments, the segment length might be a little longer
        than `segment_length`, like `segment_length` * 1.25
      reference_length_difference:
        Because of the insertion or deletion errors, the reference sequence might be
        shorter or longer than the query, so the reference segment length can be from
        `len(query) * (1 - reference_length_difference / 2)` to
        `len(query) * (1 + reference_length_difference / 2)`.
    """
    # Firstly we will find the longest range that satisfies
    # 1 - reference_length_difference / 2 <= len(reference) / len(query)
    # <= 1 + reference_length_difference / 2
    prev_break_point = 0  # index in matched_points
    candidate_ranges: List[
        Tuple[int, int]
    ] = []  # start, end indexes in matched_points
    for i in range(1, len(matched_points)):
        # The query and reference distance between current matched_point and
        # preceding matched_point.
        gap = (
            matched_points[i][0] - matched_points[i - 1][0],
            matched_points[i][1] - matched_points[i - 1][1],
        )
        if gap[0] < gap[1]:
            # +1 for smoothing (avoid div zero error)
            ratio = (
                matched_points[i][1] - matched_points[prev_break_point][1] + 1
            ) / (matched_points[i][0] - matched_points[prev_break_point][0] + 1)
            # break on current matched_point, we will finally choose the longest
            # segment.
            half = reference_length_difference / 2
            if ratio < 1 - half or ratio > 1 + half:
                candidate_ranges.append((prev_break_point, i))
                prev_break_point = i

    candidate_ranges.append((prev_break_point, len(matched_points)))

    # Find the range containing the longest query sequence
    # max_item contains the matched points we choose :
    # matched_points[max_item[0], max_item[1]]
    max_length = -1
    max_item = (0, len(matched_points))
    for c in candidate_ranges:
        current_length = matched_points[c[1] - 1][0] - matched_points[c[0]][0]
        if current_length > max_length:
            max_length = current_length
            max_item = c

    # start, end indexes of query and reference in sourced_text
    # [(query_start, query_end, target_start, target_end)]
    segments: List[Tuple[int, int, int, int]] = []

    target_doc_id = sourced_text.doc[matched_points[max_item[0]][1]]
    target_base = sourced_text.doc_splits[target_doc_id]
    next_target_base = sourced_text.doc_splits[target_doc_id + 1]

    query_doc_id = sourced_text.doc[matched_points[max_item[0]][0]]
    query_base = sourced_text.doc_splits[query_doc_id]
    next_query_base = sourced_text.doc_splits[query_doc_id + 1]

    # indexes of query and reference in sourced_text
    # Guarantee the target index is in the same reference document
    prev_target = matched_points[max_item[0]][1] - (
        matched_points[max_item[0]][0] - query_base
    )
    prev_break_point = (
        query_base,
        prev_target if prev_target >= target_base else target_base,
    )

    # Break the chosen matched_points into smaller segment, we might need to
    # extend on both side, as the first and last chosen points are sometimes
    # not the real start point and end point of query.
    for ind in range(max_item[0], max_item[1]):
        if matched_points[ind][0] - prev_break_point[0] > segment_length:
            if ind == max_item[0]:
                segments.append(
                    (
                        prev_break_point[0],
                        matched_points[ind][0],
                        prev_break_point[1],
                        matched_points[ind][1],
                    )
                )
                prev_break_point = matched_points[ind]
            else:
                query_start = prev_break_point[0]
                query_end = matched_points[ind - 1][0]
                target_start = prev_break_point[1]
                target_end = matched_points[ind - 1][1]

                ratio = (target_end - target_start) / (query_end - query_start)
                half = reference_length_difference / 2
                if ratio < 1 - half or ratio > 1 + half:
                    logging.debug(
                        f"Invalid ratio for segment: "
                        f"{query_start, query_end, target_start, target_end}"
                    )
                    continue

                segments.append(
                    (query_start, query_end, target_start, target_end)
                )
                prev_break_point = (query_end, target_end)

    query_start, target_start = prev_break_point
    query_end = next_query_base

    target_end = target_start + (query_end - query_start)
    target_end = (
        target_end if target_end <= next_target_base else next_target_base
    )

    if query_end - query_start < segment_length // 4:
        if segments:
            segments[-1] = (
                segments[-1][0],
                query_end,
                segments[-1][2],
                target_end,
            )
        else:
            segments.append((query_start, query_end, target_start, target_end))
    else:
        segments.append((query_start, query_end, target_start, target_end))
    return segments


def _combine_sub_alignments(
    sourced_text: SourcedText,
    sub_alignments: List[Dict[str, Any]],
    num_queries: int,
) -> List[Tuple[Tuple[int, int], List[Dict[str, Any]]]]:
    """
    Combine the segments broken by `_break_query` together to get a long query.

    Note: The segments here contain the levenshtein alignment.

    Args:
      sourced_text:
        The SourcedText containing the queries and references.
      sub_alignments:
        A list of segments, each item (returned by levenshtein_worker) contains
        the segment it belongs to and the levenshtein alignment.
      num_queries:
        The number of queries, this argument is for efficiency, if we know number
        of queries we can preallocate the list.
    Returns:
      Return a list of tuple containing
      ((query_start, target_start), [alignment item]).
      The `query_start` and `target_start` are indexes in sourced_text,
      the `alignment item` is a list containing `{"ref": ref, "hyp": hyp,
      "ref_pos": ref_pos, "hyp_pos": hyp_pos, "hyp_time": hyp_time}`, `ref` is
      the token from reference, `hyp` is the token from query, `ref_pos` is
      local index in reference document, `hyp_pos` is local index in query
      document, `hyp_time` is the timestamp of `hyp`.
    """
    # The container to store the returning results.
    alignments = [None] * num_queries
    prev_target_end = 0
    for sub in sub_alignments:
        if sub["alignment"][0] == -1:
            logging.warning(f"Skipping empty sub segment.")
            continue
        # for the global mode of levenshtein, there is only one alignment
        # See docs in `levenshtein_distance` for the meaning of target_align_start,
        # target_align_end and align_str
        target_align_start, target_align_end, align_str = sub["alignment"][1][0]
        query_start, query_end, target_start, target_end = sub["segment"]
        # we are doing levenshtein_distance in "global" mode
        assert target_align_start == 0, target_align_start
        assert target_end == (target_start + target_align_end + 1), (
            target_end,
            target_start + target_align_end + 1,
        )

        query_doc_id = sourced_text.doc[query_start]

        if alignments[query_doc_id] is None:
            alignments[query_doc_id] = (
                (query_start, target_start + target_align_start),
                [],
            )

        query_base = sourced_text.doc_splits[query_doc_id]
        query_source = sourced_text.sources[query_doc_id]

        # aligns : [{"ref": , "hyp": , "ref_pos": , "hyp_pos": , "hyp_time":}]
        aligns = alignments[query_doc_id][1]
        # Note: query_source could be TextSource or Transcript, only Transcript
        # has times.
        times = (
            query_source.times if isinstance(query_source, Transcript) else None
        )
        # The times is in byte level.
        time_stride = 1 if query_source.binary_text.dtype == np.uint8 else 4

        query_base_next = sourced_text.doc_splits[query_doc_id + 1]
        query_length = query_base_next - query_base
        query_index = query_start
        target_index = target_start + target_align_start

        for ali in align_str:
            query_index = (
                query_index if query_index < query_end else query_end - 1
            )
            hyp_pos = int(sourced_text.pos[query_index])

            hyp_time = (
                0 if times is None else float(times[hyp_pos * time_stride])
            )

            target_index = (
                target_index if target_index < target_end else target_end - 1
            )
            ref_pos = int(sourced_text.pos[target_index])
            if ali == "I":
                ref = ""
                hyp = chr(sourced_text.binary_text[query_index])
                query_index += 1
            elif ali == "D":
                ref = chr(sourced_text.binary_text[target_index])
                hyp = ""
                target_index += 1
            else:
                assert ali == "C" or ali == "S"
                ref = chr(sourced_text.binary_text[target_index])
                hyp = chr(sourced_text.binary_text[query_index])
                # The following two asserts guarantee the correctness of
                # levenshtein alignment. So *DO NOT* delete it.
                if ali == "C":
                    assert ref == hyp
                else:
                    assert ref != hyp
                query_index += 1
                target_index += 1
            aligns.append(
                {
                    "ref": ref,
                    "hyp": hyp,
                    "ref_pos": ref_pos,
                    "hyp_pos": hyp_pos,
                    "hyp_time": hyp_time,
                }
            )
    return alignments


def align_queries(
    sourced_text: SourcedText,
    num_query_tokens: int,
    num_close_matches: int = 2,
    segment_length: int = 5000,
    reference_length_difference: float = 0.1,
    min_matched_query_ratio: float = 0.33,
    thread_pool: Optional[ThreadPool] = None,
) -> List[Tuple[Tuple[int, int], List[Dict[str, Any]]]]:
    """
    Get levenshtein alignment for each query document in sourced_text.

    For each query document we will locate its corresponding segment in the
    reference document and return the levenshtein alignment between the query
    and its corresponding reference segment.

    Args:
      sourced_text:
        The SourcedText containing the queries and references, the first N
        documents are queries and the others are references.
      num_query_tokens:
        The number of query tokens in sourced_text, it tells the boundary between
        queries and references sourced_text[0:num_query_tokens] are query tokens,
        sourced_text[num_query_tokens:] are reference tokens.
      num_close_matches:
        The number of close_matches for each query token.
      segment_length:
        The query length might be long, in order to accelerate the levenshtein
        algorithm we will split the long query into smaller segments, `segment_length`
        is the expected length of the smaller segments.
      reference_length_difference:
        Because of the insertion or deletion errors, the reference sequence might be
        shorter or longer than the query, so the reference segment length can be from
        `len(query) * (1 - reference_length_difference / 2)` to
        `len(query) * (1 + reference_length_difference / 2)`.
      min_matched_query_ratio:
        The minimum ratio of matched query tokens, if the ratio is less than
        `min_matched_query_ratio` the query will be dropped.
      thread_pool:
        The ThreadPool to run levenshtein.
    Returns:
      Return a list of tuple containing
      ((query_start, target_start), [alignment item]).
      The `query_start` and `target_start` are indexes in sourced_text,
      the `alignment item` is a list containing `{"ref": ref, "hyp": hyp,
      "ref_pos": ref_pos, "hyp_pos": hyp_pos, "hyp_time": hyp_time}`, `ref` is
      the token from reference, `hyp` is the token from query, `ref_pos` is
      local index in reference document, `hyp_pos` is local index in query
      document, `hyp_time` is the timestamp of `hyp`.
    """

    logging.debug(
        f"Creating suffix array on a sequence with length : "
        f"{sourced_text.binary_text.size}."
    )
    suffix_array = create_suffix_array(sourced_text.binary_text)

    logging.debug(
        f"Finding close matches for {num_query_tokens} query tokens with "
        f"num_close_matches={num_close_matches}."
    )
    close_matches = find_close_matches(
        suffix_array, num_query_tokens, num_close_matches=num_close_matches
    )

    tot_query_symbols, num_close_matches = close_matches.shape
    assert num_query_tokens == tot_query_symbols, (
        num_query_tokens,
        tot_query_symbols,
    )
    num_queries = sourced_text.doc[tot_query_symbols]

    logging.debug(f"Getting alignments for {num_queries} queries.")

    row_splits = sourced_text.doc_splits
    arguments = []
    for q in range(num_queries):
        query_start = row_splits[q]
        query_end = row_splits[q + 1]
        query_len = row_splits[q + 1] - row_splits[q]

        # seq1 contains the query index in sourced_text.
        # seq2 contains the reference index in sourced_text.
        seq1 = np.arange(query_start, query_end).reshape(-1, 1)
        seq1 = np.tile(seq1, num_close_matches).flatten()

        seq2 = close_matches[query_start:query_end, :].flatten()
        # matched_points is a list of (query index, target index), global index
        # in sourced_text
        matched_points = get_longest_increasing_pairs(seq1, seq2)

        # In the algorithm of `find_close_matches`,
        # `sourced_text.binary_text.size - 1` means no close_matches
        if len(matched_points) != 0:
            trim_pos = len(matched_points) - 1
            while (
                matched_points[trim_pos][1] == sourced_text.binary_text.size - 1
            ):
                trim_pos -= 1
            matched_points = matched_points[0:trim_pos]

        if len(matched_points) == 0:
            logging.warning(
                f"Skipping query {q}, no matched points between query and target"
                f"in close_matches."
            )
            continue

        # The following code guarantees the matched points are in the same
        # reference document. We will choose the reference document that matches
        # the most number of query tokens.
        doc_ids = sourced_text.doc[np.array([x[1] for x in matched_points])]
        # we don't really care about the real doc id, we just want to know the
        # index in matched_points, minus doc_ids[0] here for efficiency.
        doc_ids = doc_ids - doc_ids[0]
        doc_splits = row_ids_to_row_splits(doc_ids)
        max_num_matches = -1
        max_ranges = (0, len(matched_points))
        for i in range(doc_splits.size - 1):
            if doc_splits[i + 1] - doc_splits[i] < 2:
                continue
            num_matches = (
                matched_points[doc_splits[i + 1] - 1][0]
                - matched_points[doc_splits[i]][0]
            )
            if num_matches > max_num_matches:
                max_num_matches = num_matches
                max_ranges = (doc_splits[i], doc_splits[i + 1])

        if max_num_matches < min_matched_query_ratio * query_len:
            logging.warning(
                f"Skipping query {q}, less than {min_matched_query_ratio * 100}"
                f"% of query tokens matched in close_matches."
            )
            continue

        matched_points = matched_points[max_ranges[0] : max_ranges[1]]

        # Break query into short segments, so that we only need to run levenshtein
        # algorithm on short sequence (for efficiency).
        # segments is a list of (query_start, query_end, target_start, target_end)
        # they are the indexes in sourced_text.
        segments = _break_query(
            sourced_text=sourced_text,
            matched_points=matched_points,
            segment_length=segment_length,
            reference_length_difference=reference_length_difference,
        )

        # prepare arguments list for levenshtein_worker
        for seg in segments:
            arguments.append((sourced_text, seg))

    def levenshtein_worker(sourced_text, segment):
        query = sourced_text.binary_text[segment[0] : segment[1]]
        target = sourced_text.binary_text[segment[2] : segment[3]]
        # Using global matching mode here, thanks to `get_longest_increasing_pairs`
        # we have very good matching points.
        alignment = levenshtein_distance(
            query=query, target=target, mode="global"
        )
        return {
            "segment": segment,
            "alignment": alignment,
            "target": target,
        }

    pool = ThreadPool() if thread_pool is None else thread_pool
    logging.debug(f"Matching with levenshtein for {len(arguments)} segments.")
    async_results = pool.starmap_async(levenshtein_worker, arguments)
    results = async_results.get()
    logging.debug("Matching with levenshtein done.")

    alignments = _combine_sub_alignments(sourced_text, results, num_queries)

    assert sourced_text.doc[num_query_tokens] == len(alignments), (
        sourced_text.doc[num_query_tokens],
        len(alignments),
    )
    return alignments


def _get_segment_candidates(
    target_source: TextSource,
    alignment,
    silence_length_to_break: float = 0.6,  # in second
    min_duration: float = 2,  # in second
    max_duration: float = 30,  # in second
    expected_duration: Tuple[float, float] = (5, 20),  # in second
    max_error_rate: float = 0.20,
    num_of_best_position: int = 4,
) -> List[Tuple[int, int, float]]:
    """
    Split the long aligned query into smaller segments.

    First, we get a list candidate breaking positions, these positions have
    long enough preceding silence (begin a split) or succeeding silence (end of
    a split), the length of silence is controlled by `silence_length_to_break`.
    If the `target_source` has punctuations, we will only start or end of a
    split at punctuation positions that indicate the end of a sentence.

    we create scores for each candidate position in the alignment, corresponding
    to how good a position it is to begin or end a split.

     - begin a split (i.e. this is first position in a segment)
        - plus score equal to num silence seconds this
          follows, up to some limit 3 seconds, i.e. this element
          of the Transcript's time minus the previous element's time; or some
          default (3.0 sec) if this is the first element of the Transcript.
        - good if there are few errors around this point in the alignment, i.e.
          score corresponding to number of ins,del,mismatch within a certain
          region of this position.
        - good if the previous non-whitespace character was a punctuation
          character.  (Lists of whitespace and punctuation characters will
          be passed in)

     - end a split (i.e. this is the last position in a segment).
        - plus score equal to number of silence seconds follows this position,
          up to some limit 3 seconds, i.e next element of the Transcript's time
          minus this element's time; or some default (3.0 sec) if this is the
          last element of the Transcript.
        - good if there are few errors around this point in the alignment, i.e.
          score corresponding to number of ins,del,mismatch within a certain
          region of this position.
        - good if the following non-whitespace character is a punctuation character.


    We then create a rule to assign scores to potential segments. This consist of
    the begin-scores, plus the end-scores, plus:
      - Some kind of penalty related to the duration of the segment, e.g.
        infinity if it's over some `max_duration` like 30 seconds or less than a
        `min_duration` like 2 seconds; else, one that encourages a duration
        between `expected_duration` like 5 to 20 seconds.
      - A bonus for the number of matches in the alignment.
      - A penalty for the number of errors in the alignment (could multiply this by
        some scale depending how much we don't want to have errors, but some errors
        are expected due to both ASR errors and normalization differences.)

    Next, we search for a good segmentation. We define the problem as getting
    the highest-scoring set of segments that do not overlap. For each
    begin_position, find the `num_of_best_position` like 4 best-scoring
    end_positions. For each end_position, find the `num_of_best_position`
    best-scoring begin_positions. Append the preceding 2 sets of segments to get
    a list of candidate segments.

    Args:
      target_source:
        A TextSource containing the matched reference.
      alignment:
        Alignment information, one item of the returned alignments from
        `align_queries`.
      silence_length_to_break:
        A threshold for deciding the possible breaking points, if a position has
        preceding or succeeding silence length greater than this value, we will
        add it as a possible breaking point.
        Caution: Only be used when there are no punctuations in target_source.
      min_duration:
        The minimum duration (in second) allowed for a segment.
      max_duration:
        The maximum duration (in second) allowed for a segment.
      expected_duration:
        The expected duration (in second) for a segment, it is a range (a tuple
        containing lower bound and upper bound).
        Note: The values must satisfy `min_duration <= expected_duration[0]` and
        `max_duration >= expected_duration[1]`.
      max_error_rate:
        The max levenshtein distance (char level) between query and target at
        the segment area, if the segments with higher error rate than this value
        will not appear in the final result list.
      num_of_best_position:
        For each candidate breaking points, there will be several possible start
        points (if the point is the end of segment) or end points (if the point
        is the start of a segment), this is the number of possible points.
        Normally this does not affect the final result too much, just leave it
        as default is OK.

    Returns:
      Returns a list of tuple, each tuple contains the start position,
      end position and score of current segment, start position and end
      position are indexes in aligns.
    """
    (query_start, target_start), aligns = alignment

    # [(index, score)]
    begin_scores: List[Tuple[int, float]] = []
    end_scores: List[Tuple[int, float]] = []

    # levenshtein errors in given region size
    half_region_size: int = 20

    errors_in_region: int = sum(
        [
            1 if align["ref"] != align["hyp"] else 0
            for align in aligns[0:half_region_size]
        ]
    )

    # will be used as max_silence, punctuation_score, init_duration_score
    # we need to make these scores similar to avoiding one of them dominating
    # the total score. Any value is ok, choose 3 here mainly based on the
    # silence length.
    base_score = 3
    max_silence = base_score  # seconds
    # score for preceding (begin) and succeeding (end) punctuation
    punctuation_score = base_score

    # Use cumsum to get number of matches and errors in a range efficiently
    cumsum_match = [0] * len(aligns)
    cumsum_error = [0] * len(aligns)

    # to avoid breaking at somethings like Mr. Mrs. etc.
    period_patterns = regex.compile(
        "(?<!Mr|Mrs|Dr|Ms|Prof|Pro|Capt|Gen|Sen|Rev|Hon|St)\."
    )
    # the largest length of the patterns.
    period_pattern_length = 5

    for i, align in enumerate(aligns):
        matched = align["ref"] == align["hyp"]
        cumsum_match[i] = (
            int(matched) if i == 0 else (cumsum_match[i - 1] + int(matched))
        )
        cumsum_error[i] = (
            int(not matched)
            if i == 0
            else (cumsum_error[i - 1] + int(not matched))
        )

        # calculate the errors in a region with a sliding window.
        if i - half_region_size >= 0:
            if (
                aligns[i - half_region_size]["ref"]
                != aligns[i - half_region_size]["hyp"]
            ):
                errors_in_region -= 1
        if i + half_region_size < len(aligns):
            if (
                aligns[i + half_region_size]["ref"]
                != aligns[i + half_region_size]["hyp"]
            ):
                errors_in_region += 1

        # silence
        prev_silence = (
            max_silence
            if i == 0
            else (align["hyp_time"] - aligns[i - 1]["hyp_time"])
        )
        prev_silence = (
            max_silence if prev_silence > max_silence else prev_silence
        )
        succ_silence = (
            max_silence
            if i == len(aligns) - 1
            else (aligns[i + 1]["hyp_time"] - align["hyp_time"])
        )
        succ_silence = (
            max_silence if succ_silence > max_silence else succ_silence
        )

        # punctuation
        prev_punctuation = 0
        j = align["ref_pos"] - 1
        while j >= 0:
            current_token = chr(target_source.binary_text[j])
            if is_punctuation(current_token, eos_only=True):
                tmp = "".join(
                    [
                        chr(x)
                        for x in target_source.binary_text[
                            j - period_pattern_length : j + 1
                        ]
                    ]
                )
                if current_token != "." or (
                    current_token == "."
                    and period_patterns.search(tmp) is not None
                ):
                    prev_punctuation = punctuation_score
                    break
                else:
                    j -= 1
            elif current_token == " " or is_punctuation(current_token):
                j -= 1
            else:
                break

        succ_punctuation = 0
        j = align["ref_pos"] + 1
        while j < target_source.binary_text.size:
            current_token = chr(target_source.binary_text[j])
            if is_punctuation(current_token, eos_only=True):
                tmp = "".join(
                    [
                        chr(x)
                        for x in target_source.binary_text[
                            j - period_pattern_length : j + 1
                        ]
                    ]
                )
                if current_token != "." or (
                    current_token == "."
                    and period_patterns.search(tmp) is not None
                ):
                    succ_punctuation = punctuation_score
                    break
                else:
                    j += 1
            elif current_token == " " or is_punctuation(current_token):
                j += 1
            else:
                break

        begin_score = (
            prev_silence
            + prev_punctuation
            - errors_in_region / half_region_size
        )

        end_score = (
            succ_silence
            + succ_punctuation
            - errors_in_region / half_region_size
        )

        if target_source.has_punctuation:
            if prev_punctuation > 0 or i == 0:
                begin_scores.append((i, begin_score))
            if succ_punctuation > 0 or i == len(aligns) - 1:
                end_scores.append((i, end_score))
        else:
            if matched and (prev_silence >= silence_length_to_break or i == 0):
                begin_scores.append((i, begin_score))
            if matched and (
                succ_silence >= silence_length_to_break or i == len(aligns) - 1
            ):
                end_scores.append((i, end_score))

    # (start, end, score)
    begin_list: List[Tuple[int, int, float]] = []
    end_list: List[Tuple[int, int, float]] = []

    init_duration_score = (
        base_score  # duration_score for segment between 5 ~ 20 seconds
    )

    last_ind = 0
    for i, item in enumerate(begin_scores):
        # Caution: Can only be modified with heappush and heappop, it is used as
        # the container of a heap.
        item_q = []
        ind = bisect_right(end_scores, item, lo=last_ind)
        last_ind = ind

        while True:
            if ind >= len(end_scores) or ind < 0:
                break

            duration = (
                aligns[end_scores[ind][0]]["hyp_time"]
                - aligns[item[0]]["hyp_time"]
            )
            if duration <= min_duration:
                ind += 1
                continue

            if duration > max_duration:
                break

            point_score = begin_scores[i][1] + end_scores[ind][1]

            # matching scores
            matched_score = (
                base_score
                * (cumsum_match[end_scores[ind][0]] - cumsum_match[item[0]])
                / (end_scores[ind][0] - item[0])
            )

            # errors penalties
            total_errors = (
                cumsum_error[end_scores[ind][0]] - cumsum_error[item[0]]
            )
            # skipping segment with too much matching errors.
            if total_errors >= (end_scores[ind][0] - item[0]) * max_error_rate:
                ind += 1
                continue
            error_score = (
                base_score * (total_errors) / (end_scores[ind][0] - item[0])
            )

            duration_score = init_duration_score

            duration_score = (
                duration_score
                - (duration - min_duration)
                / (expected_duration[0] - min_duration)
                * init_duration_score
                if duration < expected_duration[0]
                else duration_score
            )

            duration_score = (
                duration_score
                - (max_duration - duration)
                / (max_duration - expected_duration[1])
                * init_duration_score
                if duration > expected_duration[1]
                else duration_score
            )

            heappush(
                item_q,
                (
                    point_score + matched_score - error_score + duration_score,
                    (item[0], end_scores[ind][0]),
                ),
            )
            if len(item_q) > num_of_best_position:
                heappop(item_q)
            ind += 1
        while item_q:
            x = heappop(item_q)
            begin_list.append((x[1][0], x[1][1], x[0]))

    last_ind = 0
    for i, item in enumerate(end_scores):
        # Caution: Can only be modified with heappush and heappop, it is used as
        # the container of a heap.
        item_q = []
        ind = bisect_left(begin_scores, item, lo=last_ind)
        last_ind = ind

        while True:
            if ind < 0 or ind >= len(begin_scores):
                break

            duration = (
                aligns[item[0]]["hyp_time"]
                - aligns[begin_scores[ind][0]]["hyp_time"]
            )

            if duration <= min_duration:
                ind -= 1
                continue

            if duration >= max_duration:
                break

            point_score = begin_scores[ind][1] + end_scores[i][1]
            # matching scores
            matched_score = (
                base_score
                * (cumsum_match[item[0]] - cumsum_match[begin_scores[ind][0]])
                / (item[0] - begin_scores[ind][0])
            )

            # errors penalties
            total_errors = (
                cumsum_error[item[0]] - cumsum_error[begin_scores[ind][0]]
            )
            # skipping segment with too much matching errors.
            if (
                total_errors
                >= (item[0] - begin_scores[ind][0]) * max_error_rate
            ):
                ind -= 1
                continue
            error_score = (
                base_score * (total_errors) / (item[0] - begin_scores[ind][0])
            )

            duration_score = init_duration_score

            duration_score = (
                duration_score
                - (duration - min_duration)
                / (expected_duration[0] - min_duration)
                * init_duration_score
                if duration < expected_duration[0]
                else duration_score
            )

            duration_score = (
                duration_score
                - (max_duration - duration)
                / (max_duration - expected_duration[1])
                * init_duration_score
                if duration > expected_duration[1]
                else duration_score
            )

            heappush(
                item_q,
                (
                    point_score + matched_score - error_score + duration_score,
                    (begin_scores[ind][0], item[0]),
                ),
            )
            if len(item_q) > num_of_best_position:
                heappop(item_q)
            ind -= 1
        while item_q:
            x = heappop(item_q)
            end_list.append((x[1][0], x[1][1], x[0]))

    candidates = begin_list + end_list
    return candidates


def _split_into_segments(
    query_source: Union[Transcript, TextSource],
    target_source: TextSource,
    alignment: Tuple[Tuple[int, int], List[Dict[str, Any]]],
    preceding_context_length: int = 1000,
    timestamp_position: str = "middle",  # previous, middle, current
    silence_length_to_break: float = 0.6,  # in second
    overlap_ratio: float = 0.35,  # percentage
    min_duration: float = 2,  # in second
    max_duration: float = 30,  # in second
    expected_duration: Tuple[float, float] = (5, 20),  # in second
    max_error_rate: float = 0.15,
    num_of_best_position: int = 4,
) -> List[Dict[str, Union[str, int, float]]]:
    """
    Split a long aligned query into smaller segments.

    We will create scores for each position in the alignment, corresponding to
    how good a position it is to begin or end a split. We can then create a rule
    to assign scores to potential segments. The scores would consist of the
    begin-scores, plus the end-scores, plus some kind of scores of given segment
    (duration, matching errors .etc).

    Args:
      query_source:
        An instance of Transcript or TextSource containing the query.
      target_source:
        An instance of TextSource containing the matched reference.
      alignment:
        The alignment, an item in the list returned by `align_queries`.
      preceding_context_length:
        The number of characters of preceding context.
      timestamp_position:
        It indicates which token we will get `start_time` from, valid values are
        `previous`, `middle` and `current`. If it equals to `current` the
        `start_time` is the timestamp of token in `begin_pos`, if it equals to
        `previous` the `start_time` is the timestamp of token in `begin_pos - 1`
        if it equals to `middle`, the `start_time` is the averaged timestamp of
        tokens in `begin_pos` and `begin_pos - 1`.
      silence_length_to_break:
        A threshold for deciding the possible breaking points, if a position has
        preceding or succeeding silence length greater than this value, we will
        add it as a possible breaking point.
        Caution: Only be used when there are no punctuations in target_source.
      overlap_ratio:
        The ratio of overlapping part to the query or existing segments. If the
        ratio is greater than `overlap_ratio` we will drop the query or existing
        segment.
      min_duration:
        The minimum duration (in second) allowed for a segment.
      max_duration:
        The maximum duration (in second) allowed for a segment.
      expected_duration:
        The expected duration (in second) for a segment, it is a range (a tuple
        containing lower bound and upper bound).
        Note: The values must satisfy `min_duration <= expected_duration[0]` and
        `max_duration >= expected_duration[1]`.
      max_error_rate:
        The max levenshtein distance (char level) between query and target at
        the segment area, if the segments with higher error rate than this value
        will not appear in the final result list.
      num_of_best_position:
        For each candidate breaking points, there will be several possible start
        points (if the point is the end of segment) or end points (if the point
        is the start of a segment), this is the number of possible points.
        Normally this does not affect the final result too much, just leave it
        as default is OK.

    Return
      Returns a list of Dict containing the details of each segment, looks like

         {
             "begin_byte": int,   # begin position in original target source
             "end_byte": int,     # end position in original target source
             "start_time": float, # start timestamp in the audio
             "duration": float,   # duration of this segment
             "hyp": str,          # text from query source
             "ref": str,          # text from target source
             "pre_ref": str,      # preceding text from target source
             "pre_hyp": str,      # preceding text from query source
             "post_ref": str,     # succeeding text from target source
             "post_hyp": str,     # succeeding text from query source
         }
    """
    (query_start, target_start), aligns = alignment

    # candidates : (start, end, score), start and end are indexes in aligns
    candidates = _get_segment_candidates(
        target_source=target_source,
        alignment=alignment,
        silence_length_to_break=silence_length_to_break,
        min_duration=min_duration,
        max_duration=max_duration,
        expected_duration=expected_duration,
        max_error_rate=max_error_rate,
        num_of_best_position=num_of_best_position,
    )

    candidates = sorted(candidates, key=lambda x: x[2], reverse=True)

    # Handle the overlapping
    # Caution: Don't modified selected_ranges, it will be manipulated in
    # `is_overlap` and will be always kept sorted.
    # Don't modified selected_indexes also, it will be manipulated in `is_overlap`
    # according to selected_ranges.
    selected_ranges: List[Tuple[float, float]] = []
    selected_indexes: List[int] = []
    segments = []
    overlapped_segments = []
    for r in candidates:
        status, index = is_overlap(
            selected_ranges,
            selected_indexes,
            query=(aligns[r[0]]["hyp_time"], aligns[r[1]]["hyp_time"]),
            segment_index=len(segments),
            overlap_ratio=overlap_ratio,
        )
        if status:
            if index is not None:
                overlapped_segments.append(index)
                segments.append(r)
        else:
            segments.append(r)
    for index in sorted(overlapped_segments, reverse=True):
        segments.pop(index)

    results = []

    for seg in segments:
        begin_pos = aligns[seg[0]]["ref_pos"]
        # trim the prefix space
        while chr(target_source.binary_text[begin_pos]) == " ":
            begin_pos += 1

        while begin_pos >= 1:
            current_token = chr(target_source.binary_text[begin_pos - 1])
            if current_token in PUCTUATIONS["left"]:
                begin_pos -= 1
            else:
                break

        end_pos = aligns[seg[1]]["ref_pos"]
        while end_pos + 1 < target_source.binary_text.size:
            current_token = chr(target_source.binary_text[end_pos + 1])
            if (
                current_token in PUCTUATIONS["right"]
                or current_token in PUCTUATIONS["eos"]
            ):
                end_pos += 1
            else:
                break

        preceding_index = seg[0] if seg[0] == 0 else seg[0] - 1
        succeeding_index = seg[1] if seg[1] == len(aligns) - 1 else seg[1] + 1

        if timestamp_position == "middle":
            start_time = (
                aligns[preceding_index]["hyp_time"] + aligns[seg[0]]["hyp_time"]
            ) / 2
            end_time = (
                aligns[succeeding_index]["hyp_time"]
                + aligns[seg[1]]["hyp_time"]
            ) / 2
        elif timestamp_position == "previous":
            start_time = aligns[preceding_index]["hyp_time"]
            end_time = aligns[seg[1]]["hyp_time"]
        else:
            assert timestamp_position == "current", (
                timestamp_position,
                "current",
            )
            start_time = aligns[seg[0]]["hyp_time"]
            end_time = aligns[seg[1]]["hyp_time"]

        hyp_begin_pos = aligns[seg[0]]["hyp_pos"]
        while chr(query_source.binary_text[hyp_begin_pos]) == " ":
            hyp_begin_pos += 1

        hyp_end_pos = aligns[seg[1]]["hyp_pos"]
        hyp = "".join(
            [
                chr(i)
                for i in query_source.binary_text[
                    hyp_begin_pos : hyp_end_pos + 1
                ]
            ]
        )

        # output one more token for reference to include the possible punctuation.
        # end_pos plus 1 here is safe, it is probably a space or punctuation.
        ref = "".join(
            [chr(i) for i in target_source.binary_text[begin_pos : end_pos + 1]]
        )

        preceding_pos = (
            begin_pos - preceding_context_length
            if begin_pos - preceding_context_length >= 0
            else 0
        )
        preceding_ref = "".join(
            [chr(i) for i in target_source.binary_text[preceding_pos:begin_pos]]
        )
        succeeding_ref = "".join(
            [
                chr(i)
                for i in target_source.binary_text[
                    end_pos + 1 : end_pos + preceding_context_length
                ]
            ]
        )

        preceding_pos = (
            hyp_begin_pos - preceding_context_length
            if hyp_begin_pos - preceding_context_length > 0
            else 0
        )
        preceding_hyp = "".join(
            [
                chr(i)
                for i in query_source.binary_text[preceding_pos:hyp_begin_pos]
            ]
        )
        succeeding_hyp = "".join(
            [
                chr(i)
                for i in query_source.binary_text[
                    hyp_end_pos + 1 : hyp_begin_pos + preceding_context_length
                ]
            ]
        )

        results.append(
            {
                "begin_byte": begin_pos,
                "end_byte": end_pos,
                "start_time": start_time,
                "duration": math.floor(1000 * (end_time - start_time)) / 1000,
                "hyp": hyp,
                "ref": ref,
                "pre_ref": preceding_ref,
                "pre_hyp": preceding_hyp,
                "post_ref": succeeding_ref,
                "post_hyp": succeeding_hyp,
            }
        )
    return results


def _split_helper(
    query_source: Union[TextSource, Transcript],
    target_source: TextSource,
    cut_index: Tuple[int, int],
    alignment: Tuple[Tuple[int, int], List[Dict[str, Any]]],
    preceding_context_length: int,
    timestamp_position: str,
    silence_length_to_break: float,
    overlap_ratio: float,
    min_duration: float,
    max_duration: float,
    expected_duration: Tuple[float, float],
    max_error_rate: float,
    num_of_best_position: int,
):
    """
    A worker function for splitting aligned query.
    """
    segments = _split_into_segments(
        query_source,
        target_source,
        alignment,
        preceding_context_length=preceding_context_length,
        timestamp_position=timestamp_position,
        silence_length_to_break=silence_length_to_break,
        overlap_ratio=overlap_ratio,
        min_duration=min_duration,
        max_duration=max_duration,
        expected_duration=expected_duration,
        max_error_rate=max_error_rate,
        num_of_best_position=num_of_best_position,
    )
    return cut_index, segments


def split_aligned_queries(
    sourced_text: SourcedText,
    alignments: List[Tuple[Tuple[int, int], List[Dict[str, Any]]]],
    cut_indexes: List[Tuple[int, int]],
    process_pool: Optional[Pool] = None,
    preceding_context_length: int = 1000,
    timestamp_position: str = "current",  # previous, middle, current
    silence_length_to_break: float = 0.6,  # in second
    overlap_ratio: float = 0.35,
    min_duration: float = 2,  # in second
    max_duration: float = 30,  # in second
    expected_duration: Tuple[float, float] = (5, 20),  # in second
    max_error_rate: float = 0.15,
    num_of_best_position: int = 4,
) -> List[Dict[str, Union[str, int, float]]]:
    """
    Split the aligned queries into smaller segments (A query might have several
    hours of audio, which is not suitable for ASR training)

    Args:
      sourced_text:
        The SourcedText containing the queries and references.
      alignments:
        The alignments returned by function align_queries. The length of it is
        equals to the number of queries.
      cut_indexes:
        A list of tuple containing the original cut index and supervision index
        of the query([(cut index, sup index)]), it satisfies
        `len(cut_indexes) == len(alignments)`
      process_pool:
        The process pool to split aligned queries. The algorithms are
        implemented in python, so we use process pool to get rid of the effect
        of GIL.
      preceding_context_length:
        The number of characters of preceding context.
      timestamp_position:
        It indicates which token we will get `start_time` from, valid values are
        `previous`, `middle` and `current`. If it equals to `current` the
        `start_time` is the timestamp of token in `begin_pos`, if it equals to
        `previous` the `start_time` is the timestamp of token in `begin_pos - 1`
        if it equals to `middle`, the `start_time` is the averaged timestamp of
        tokens in `begin_pos` and `begin_pos - 1`.
      silence_length_to_break:
        A threshold for deciding the possible breaking points, if a position has
        preceding or succeeding silence length greater than this value, we will
        add it as a possible breaking point.
        Caution: Only be used when there are no punctuations in target_source.
      overlap_ratio:
        The ratio of overlapping part to the query or existing segments. If the
        ratio is greater than `overlap_ratio` we will drop the query or existing
        segment.
      min_duration:
        The minimum duration (in second) allowed for a segment.
      max_duration:
        The maximum duration (in second) allowed for a segment.
      expected_duration:
        The expected duration (in second) for a segment, it is a range (a tuple
        containing lower bound and upper bound).
        Note: The values must satisfy `min_duration <= expected_duration[0]` and
        `max_duration >= expected_duration[1]`.
      max_error_rate:
        The max levenshtein distance (char level) between query and target at
        the segment area, if the segments with higher error rate than this value
        will not appear in the final result list.
      num_of_best_position:
        For each candidate breaking points, there will be several possible start
        points (if the point is the end of segment) or end points (if the point
        is the start of a segment), this is the number of possible points.
        Normally this does not affect the final result too much, just leave it
        as default is OK.

    Return:
      Returns a list of Dict containing the details of each segment, looks like

         {
             "begin_byte": int,   # begin position in original target source
             "end_byte": int,     # end position in original target source
             "start_time": float, # start timestamp in the audio
             "duration": float,   # duration of this segment
             "hyp": str,          # text from query source
             "ref": str,          # text from target source
             "pre_ref": str,      # preceding text from target source
             "pre_hyp": str,      # preceding text from query source
             "post_ref": str,     # succeeding text from target source
             "post_hyp": str,     # succeeding text from query source
         }

    """
    arguments = []
    aligned_length = 0
    for i in range(len(alignments)):
        if alignments[i] is not None:
            (query_start, target_start), aligns = alignments[i]
            aligned_length += aligns[-1]["hyp_time"] - aligns[0]["hyp_time"]
            query_source = sourced_text.sources[sourced_text.doc[query_start]]
            target_source = sourced_text.sources[sourced_text.doc[target_start]]
            arguments.append(
                (
                    query_source,
                    target_source,
                    cut_indexes[i],
                    alignments[i],
                    preceding_context_length,
                    timestamp_position,
                    silence_length_to_break,
                    overlap_ratio,
                    min_duration,
                    max_duration,
                    expected_duration,
                    max_error_rate,
                    num_of_best_position,
                )
            )
    logging.debug(
        f"Aligned length : {aligned_length} seconds for {len(arguments)} queries."
    )

    logging.debug(f"Splitting into segments for {len(arguments)} queries.")
    pool = Pool() if process_pool is None else process_pool
    async_results = pool.starmap_async(_split_helper, arguments)
    results = async_results.get()
    logging.debug(f"Splitting into segments done.")

    return results



import os
import logging
from bisect import bisect_left
from typing import List, Optional, Tuple, Union
from datetime import datetime
from pathlib import Path
import numpy as np
from _textsearch import row_ids_to_row_splits as _row_ids_to_row_splits

Pathlike = Union[str, Path]

PUCTUATIONS = {
    "all": set("'.;?!():-<>/\"。；？！（）：-《》【】”“"),
    "eos": set(".?,，!。？！"),
    "left": set("\"'(<《【“"),
    "right": set("\"')>》】”"),
}


class AttributeDict(dict):
    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError(f"No such attribute '{key}'")

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        if key in self:
            del self[key]
            return
        raise AttributeError(f"No such attribute '{key}'")


def setup_logger(
    log_filename: Pathlike,
    log_level: str = "info",
    dist: Optional[Tuple[int, int]] = None,
    use_console: bool = True,
) -> None:
    """Setup log level.

    Args:
      log_filename:
        The filename to save the log.
      log_level:
        The log level to use, e.g., "debug", "info", "warning", "error",
        "critical"
      use_console:
        True to also print logs to console.
    """
    now = datetime.now()
    date_time = now.strftime("%Y-%m-%d-%H-%M-%S")
    if dist is not None:
        rank, world_size = dist
        formatter = f"%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] ({rank}/{world_size}) %(message)s"  # noqa
        log_filename = f"{log_filename}-{date_time}-{rank}"
    else:
        formatter = (
            "%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"
        )
        log_filename = f"{log_filename}-{date_time}"

    os.makedirs(os.path.dirname(log_filename), exist_ok=True)

    level = logging.ERROR
    if log_level == "debug":
        level = logging.DEBUG
    elif log_level == "info":
        level = logging.INFO
    elif log_level == "warning":
        level = logging.WARNING
    elif log_level == "critical":
        level = logging.CRITICAL

    logging.basicConfig(
        filename=log_filename,
        format=formatter,
        level=level,
        filemode="w",
    )
    if use_console:
        console = logging.StreamHandler()
        console.setLevel(level)
        console.setFormatter(logging.Formatter(formatter))
        logging.getLogger("").addHandler(console)


def row_ids_to_row_splits(row_ids: np.ndarray) -> np.ndarray:
    """Convert row ids to row splits.

    Args:
      row_ids:
        A 1-D array of dtype np.uint32 containing row ids.
    Returns:
      Return a 1-D array of dtype np.uint32 containing row splits.
    """
    assert row_ids.ndim == 1, row_ids.ndim
    assert row_ids.dtype == np.uint32, row_ids.dtype

    row_ids = np.ascontiguousarray(row_ids)
    num_rows = row_ids[-1] + 1
    row_splits = np.empty(num_rows + 1, dtype=np.uint32)

    _row_ids_to_row_splits(row_ids, row_splits)
    return row_splits


def is_overlap(
    ranges: List[Tuple[float, float]],
    indexes: List[int],
    query: Tuple[float, float],
    segment_index: int,
    overlap_ratio: float = 0.25,
) -> Tuple[bool, Union[int, None]]:
    """
    Return True if the given range overlaps with the existing ranges.

    Caution:
      `ranges` and `indexes` will be modified in this function.

    Note: overlapping here means the length of overlapping area is greater than
    some threshold (currently, the threshold is `overlap_ratio` multiply the length
    of the query or existing ranges).

    Args:
      ranges:
        The existing ranges, it is sorted in ascending order on input, and we will
        keep it sorted in this function.
      indexes:
        The index (into the selected segments) of each range belongs to.
      query:
        The given range.
      segment_index:
        The index (into the selected segments) of query to be inserted.
      overlap_ratio:
        The ratio of overlapping part to the query or existing segments. If the
        ratio is greater than `overlap_ratio` we will drop the query or existing
        segment.

    Return:
      Return (False, None) if no overlapping between query and existing ranges.
      Return (True, None) if the ratio of overlapping part to query is greater
      than `overlap_ratio`.
      Return (True, dindex) if the ratio of overlapping part to existing range
      is greater than `overlap_ratio`, `dindex` is the index (can get from indexes)
      of the existing range.
    """
    index = bisect_left(ranges, query)
    if not ranges:
        ranges.insert(index, query)
        indexes.insert(index, segment_index)
        return False, None

    # overlapping on query
    if index > 0:
        is_overlap = (
            ranges[index - 1][1] - query[0]
            > (query[1] - query[0]) * overlap_ratio
        )
        if is_overlap:
            return True, None

    if index < len(ranges):
        is_overlap = (
            query[1] - ranges[index][0] > (query[1] - query[0]) * overlap_ratio
        )
        if is_overlap:
            return True, None

    # overlapping on existing ranges
    is_overlap_left = False
    if index > 0:
        is_overlap_left = (
            ranges[index - 1][1] - query[0]
            > (ranges[index - 1][1] - ranges[index - 1][0]) * overlap_ratio
        )

    is_overlap_right = False
    if index < len(ranges):
        is_overlap_right = (
            query[1] - ranges[index][0]
            > (ranges[index][1] - ranges[index][0]) * overlap_ratio
        )

    if is_overlap_left or is_overlap_right:
        if is_overlap_left and not is_overlap_right:
            ranges.insert(index, query)
            ranges.pop(index - 1)
            indexes.insert(index, segment_index)
            dindex = indexes.pop(index - 1)
            return True, dindex
        elif is_overlap_right and not is_overlap_left:
            ranges.insert(index, query)
            ranges.pop(index + 1)
            indexes.insert(index, segment_index)
            dindex = indexes.pop(index + 1)
            return True, dindex
        else:
            return True, None
    else:
        ranges.insert(index, query)
        indexes.insert(index, segment_index)
        return False, None


def is_punctuation(c: str, eos_only: bool = False) -> bool:
    """
    Return True if the given character is a punctuation.

    Args:
      c:
        The given character.
      eos_only:
        If True the punctuations are only those indicating end of a sentence (,.?! for now).
    """
    if eos_only:
        return c in PUCTUATIONS["eos"]
    return c in PUCTUATIONS["all"]


def str2bool(v):
    """Used in argparse.ArgumentParser.add_argument to indicate
    that a type is a bool type and user can enter

        - yes, true, t, y, 1, to represent True
        - no, false, f, n, 0, to represent False

    See https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse  # noqa
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")



#!/usr/bin/env python3

import unittest

import numpy as np

from textsearch import TextSource


class TestTextSource(unittest.TestCase):
    def test_use_utf8(self):
        name = "test"
        s = "zażółć gęślą jaźń\n你好Hallo"
        source = TextSource.from_str(name=name, s=s, use_utf8=True)

        assert source.name == name, (source.name, name)

        assert isinstance(source.binary_text, np.ndarray)
        assert source.binary_text.dtype == np.uint8, source.binary_text.dtype

        expected_binary_text = s.encode("utf-8")
        assert source.binary_text.tobytes() == expected_binary_text

        assert isinstance(source.text, str), type(source.text)
        assert source.text == s, (source.text, s)

        assert source.pos is None, source.pos

    def test_without_using_utf8(self):
        name = "test"
        s = "zażółć gęślą jaźń\n你好Hallo"
        source = TextSource.from_str(name=name, s=s, use_utf8=False)

        assert source.name == name, (source.name, name)

        assert isinstance(source.binary_text, np.ndarray)
        assert source.binary_text.dtype == np.int32, source.binary_text.dtype

        expected_binary_text = np.fromiter((ord(i) for i in s), dtype=np.int32)
        np.testing.assert_equal(source.binary_text, expected_binary_text)

        assert isinstance(source.text, str), type(source.text)
        assert source.text == s, (source.text, s)

        assert isinstance(source.pos, np.ndarray), type(source.pos)
        assert source.pos.dtype == np.uint32, source.pos.dtype
        # fmt: off
        # A Polish character occupies 2 bytes in utf-8 while
        # a Chinese character occupies 3 bytes
        np.testing.assert_equal(
            source.pos,
            np.array(
                [0, 1, 2, 4, 6, 8, 10, 11, 12, 14, 16, 17, 19, 20, 21, 22, 24,
                 26, 27, 30, 33, 34, 35, 36, 37], dtype=np.uint32
            ),
        )
        # fmt: on


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import unittest

import numpy as np

from textsearch import TextSource, Transcript


class TestTranscript(unittest.TestCase):
    def test_use_utf8(self):
        text = "▁I ▁saw ▁a ▁girl ▁with ▁a ▁ te le s c o pe".split()
        begin_times = []
        for i in range(len(text)):
            begin_times.append(i * 0.25)
        d = {
            "text": text,
            "begin_times": begin_times,
        }
        transcript = Transcript.from_dict(name="test", d=d, use_utf8=True)
        source = TextSource.from_str(name="test", s="".join(text), use_utf8=True)

        assert transcript.binary_text.dtype == np.uint8, transcript.binary_text.dtype
        np.testing.assert_equal(transcript.binary_text, source.binary_text)

        assert transcript.binary_text.size == transcript.times.size, (
            transcript.binary_text.size,
            transcript.time.size,
        )

        last_pos = 0
        for t, time in zip(text, begin_times):
            num_bytes = len(t.encode("utf-8"))
            np.testing.assert_equal(
                transcript.times[last_pos : last_pos + num_bytes], time
            )
            last_pos += num_bytes

    def test_without_using_utf8(self):
        text = "▁I ▁saw ▁a ▁girl ▁with ▁a ▁ te le s c o pe".split()
        begin_times = []
        for i in range(len(text)):
            begin_times.append(i * 0.25)
        d = {
            "text": text,
            "begin_times": begin_times,
        }
        transcript = Transcript.from_dict(name="test", d=d, use_utf8=False)
        source = TextSource.from_str(name="test", s="".join(text), use_utf8=False)

        assert transcript.binary_text.dtype == np.int32, transcript.binary_text.dtype

        np.testing.assert_equal(transcript.binary_text, source.binary_text)

        assert transcript.binary_text.size * 4 == transcript.times.size, (
            transcript.binary_text.size * 4,
            transcript.times.size,
        )

        last_pos = 0
        for t, time in zip(text, begin_times):
            num_bytes = len(t) * 4
            np.testing.assert_equal(
                transcript.times[last_pos : last_pos + num_bytes], time
            )
            last_pos += num_bytes


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
#
# Copyright      2023  Xiaomi Corp.       (authors: Wei Kang)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R find_close_matches_test_py

import unittest
import numpy as np

from textsearch import create_suffix_array, find_close_matches


class TestFindCloseMatches(unittest.TestCase):
    def test_find_close_matches(self):
        """
        The suffix array contains the information below, the first column is
        the index into the original texts (the only thing suffix array store),
        the following column tell the suffix and which type (query or reference)
        this token belongs to.

        6, type : query, suffix : alloiholloyouyouhellome
        1, type : query, suffix : ellohalloiholloyouyouhellome
        23, type : reference, suffix : ellome
        28, type : reference, suffix : e
        5, type : query, suffix : halloiholloyouyouhellome
        0, type : query, suffix : hellohalloiholloyouyouhellome
        22, type : reference, suffix : hellome
        11, type : reference, suffix : holloyouyouhellome
        10, type : reference, suffix : iholloyouyouhellome
        2, type : query, suffix : llohalloiholloyouyouhellome
        7, type : query, suffix : lloiholloyouyouhellome
        24, type : reference, suffix : llome
        13, type : reference, suffix : lloyouyouhellome
        3, type : query, suffix : lohalloiholloyouyouhellome
        8, type : query, suffix : loiholloyouyouhellome
        25, type : reference, suffix : lome
        14, type : reference, suffix : loyouyouhellome
        27, type : reference, suffix : me
        4, type : query, suffix : ohalloiholloyouyouhellome
        9, type : query, suffix : oiholloyouyouhellome
        12, type : reference, suffix : olloyouyouhellome
        26, type : reference, suffix : ome
        20, type : reference, suffix : ouhellome
        17, type : reference, suffix : ouyouhellome
        15, type : reference, suffix : oyouyouhellome
        21, type : reference, suffix : uhellome
        18, type : reference, suffix : uyouhellome
        19, type : reference, suffix : youhellome
        16, type : reference, suffix : youyouhellome
        29, type : reference, suffix :
        """
        queries = ["hello", "hallo"]
        documents = ["iholloyou", "youhellome"]

        # texts will be : "hellohalloiholloyouyouhellome"
        texts = "".join(queries) + "".join(documents)
        texts_array = np.fromstring(texts, dtype=np.int8)
        suffix_array = create_suffix_array(texts_array)

        query_len = len("".join(queries))
        output = find_close_matches(suffix_array, query_len, num_close_matches=2)

        # Take the first token of query as an example, it is 'h', first we
        # will find the element 0 in suffix_array, from the comment lines above,
        # it's the sixth element of suffix_array, the nearest
        # reference before and after it are the fourth and seventh element, so
        # its close match references are [28, 22].
        # You can check the other tokens the same way.
        expected_output = np.array(
            [
                [28, 22],
                [28, 23],
                [10, 24],
                [13, 25],
                [27, 12],
                [28, 22],
                [28, 23],
                [10, 24],
                [13, 25],
                [27, 12],
            ],
            dtype=np.int32,
        )
        self.assertTrue((output == expected_output).all())

        output = find_close_matches(suffix_array, query_len, num_close_matches=4)
        expected_output = np.array(
            [
                [23, 28, 22, 11],
                [28, 28, 23, 28],
                [11, 10, 24, 13],
                [24, 13, 25, 14],
                [14, 27, 12, 26],
                [23, 28, 22, 11],
                [28, 28, 23, 28],
                [11, 10, 24, 13],
                [24, 13, 25, 14],
                [14, 27, 12, 26],
            ],
            dtype=np.int32,
        )
        self.assertTrue((output == expected_output).all())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import unittest

import numpy as np

from textsearch import row_ids_to_row_splits


class TestRowIdsToRowSplits(unittest.TestCase):
    def test_row_ids_to_row_splits(self):
        row_ids = np.array([0, 0, 0, 1, 1], dtype=np.uint32)
        row_splits = row_ids_to_row_splits(row_ids)
        expected_row_splits = np.array([0, 3, 5], dtype=np.uint32)
        np.testing.assert_equal(row_splits, expected_row_splits)
        assert row_splits.dtype == np.uint32, row_splits.dtype

    def test_row_ids_to_row_splits_case2(self):
        row_ids = np.array([1, 1, 3, 3, 5], dtype=np.uint32)
        row_splits = row_ids_to_row_splits(row_ids)
        print(row_splits)
        expected_row_splits = np.array([0, 0, 2, 2, 4, 4, 5], dtype=np.uint32)
        np.testing.assert_equal(row_splits, expected_row_splits)
        assert row_splits.dtype == np.uint32, row_splits.dtype


if __name__ == "__main__":
    unittest.main()
